@@ -55,9 +55,15 @@ enum lxcfs_virt_t {
 	LXC_TYPE_PROC_LOADAVG,
 #define LXC_TYPE_PROC_LOADAVG_PATH "/proc/loadavg"
 
+	LXC_TYPE_PROC_SLABINFO,
+#define LXC_TYPE_PROC_SLABINFO_PATH "/proc/slabinfo"
+
+	LXC_TYPE_SYS,
 	LXC_TYPE_SYS_DEVICES,
 	LXC_TYPE_SYS_DEVICES_SYSTEM,
 	LXC_TYPE_SYS_DEVICES_SYSTEM_CPU,
+	LXC_TYPE_SYS_DEVICES_SYSTEM_CPU_SUBDIR,
+	LXC_TYPE_SYS_DEVICES_SYSTEM_CPU_SUBFILE,
 
 	LXC_TYPE_SYS_DEVICES_SYSTEM_CPU_ONLINE,
 #define LXC_TYPE_SYS_DEVICES_SYSTEM_CPU_ONLINE_PATH "/sys/devices/system/cpu/online"
