@@ -1,29 +1,13 @@
 /* SPDX-License-Identifier: LGPL-2.1+ */
 
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE
-#endif
-
 #include "config.h"
 
-#ifdef HAVE_FUSE3
-#ifndef FUSE_USE_VERSION
-#define FUSE_USE_VERSION 30
-#endif
-#else
-#ifndef FUSE_USE_VERSION
-#define FUSE_USE_VERSION 26
-#endif
-#endif
-
-#define _FILE_OFFSET_BITS 64
-
 #include <alloca.h>
 #include <dirent.h>
 #include <dlfcn.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <fuse.h>
+#include <getopt.h>
 #include <libgen.h>
 #include <pthread.h>
 #include <sched.h>
@@ -40,6 +24,8 @@
 #include <sys/socket.h>
 #include <linux/limits.h>
 
+#include "lxcfs_fuse.h"
+
 #include "bindings.h"
 #include "lxcfs_fuse_compat.h"
 #include "macro.h"
@@ -151,7 +137,7 @@ static void do_reload(void)
 #else
         ret = snprintf(lxcfs_lib_path, sizeof(lxcfs_lib_path), "/usr/local/lib/lxcfs/liblxcfs.so");
 #endif
-	if (ret < 0 || ret >= sizeof(lxcfs_lib_path))
+	if (ret < 0 || (size_t)ret >= sizeof(lxcfs_lib_path))
 		log_exit("Failed to create path to open liblxcfs");
 
         dlopen_handle = dlopen(lxcfs_lib_path, RTLD_LAZY);
@@ -297,6 +283,22 @@ static int do_cg_write(const char *path, const char *buf, size_t size,
 	return __cg_write(path, buf, size, offset, fi);
 }
 
+static int do_sys_write(const char *path, const char *buf, size_t size,
+		       off_t offset, struct fuse_file_info *fi)
+{
+	char *error;
+	int (*__sys_write)(const char *path, const char *buf, size_t size,
+			  off_t offset, struct fuse_file_info *fi);
+
+	dlerror();
+	__sys_write = (int (*)(const char *, const char *, size_t, off_t, struct fuse_file_info *))dlsym(dlopen_handle, "sys_write");
+	error = dlerror();
+	if (error)
+		return log_error(-1, "%s - Failed to find sys_write()", error);
+
+	return __sys_write(path, buf, size, offset, fi);
+}
+
 static int do_cg_mkdir(const char *path, mode_t mode)
 {
 	char *error;
@@ -401,6 +403,19 @@ static int do_sys_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 	return __sys_readdir(path, buf, filler, offset, fi);
 }
 
+static int do_sys_readlink(const char *path, char *buf, size_t size)
+{
+	char *error;
+	int (*__sys_readlink)(const char *path, char *buf, size_t size);
+
+	dlerror();
+	__sys_readlink = (int (*)(const char *, char *, size_t))dlsym(dlopen_handle, "sys_readlink");
+	error = dlerror();
+	if (error)
+		return log_error(-1, "%s - Failed to find sys_readlink()", error);
+
+	return __sys_readlink(path, buf, size);
+}
 
 static int do_cg_open(const char *path, struct fuse_file_info *fi)
 {
@@ -472,6 +487,20 @@ static int do_sys_open(const char *path, struct fuse_file_info *fi)
 	return __sys_open(path, fi);
 }
 
+static int do_sys_opendir(const char *path, struct fuse_file_info *fi)
+{
+	char *error;
+	int (*__sys_opendir)(const char *path, struct fuse_file_info *fi);
+
+	dlerror();
+	__sys_opendir = (int (*)(const char *path, struct fuse_file_info *fi))dlsym(dlopen_handle, "sys_opendir");
+	error = dlerror();
+	if (error)
+		return log_error(-1, "%s - Failed to find sys_opendir()", error);
+
+	return __sys_opendir(path, fi);
+}
+
 static int do_sys_access(const char *path, int mode)
 {
 	char *error;
@@ -570,7 +599,7 @@ static int do_sys_releasedir(const char *path, struct fuse_file_info *fi)
 	return __sys_releasedir(path, fi);
 }
 
-#ifdef HAVE_FUSE3
+#if HAVE_FUSE3
 static int lxcfs_getattr(const char *path, struct stat *sb, struct fuse_file_info *fi)
 #else
 static int lxcfs_getattr(const char *path, struct stat *sb)
@@ -631,13 +660,17 @@ static int lxcfs_opendir(const char *path, struct fuse_file_info *fi)
 	if (strcmp(path, "/proc") == 0)
 		return 0;
 
-	if (strncmp(path, "/sys", 4) == 0)
-		return 0;
+	if (strncmp(path, "/sys", 4) == 0) {
+		up_users();
+		ret = do_sys_opendir(path, fi);
+		down_users();
+		return ret;
+	}
 
 	return -ENOENT;
 }
 
-#ifdef HAVE_FUSE3
+#if HAVE_FUSE3
 static int lxcfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 			 off_t offset, struct fuse_file_info *fi, enum fuse_readdir_flags flags)
 #else
@@ -809,6 +842,27 @@ int lxcfs_write(const char *path, const char *buf, size_t size, off_t offset,
 		return ret;
 	}
 
+	if (strncmp(path, "/sys", 4) == 0) {
+		up_users();
+		ret = do_sys_write(path, buf, size, offset, fi);
+		down_users();
+		return ret;
+	}
+
+	return -EINVAL;
+}
+
+int lxcfs_readlink(const char *path, char *buf, size_t size)
+{
+	int ret;
+
+	if (strncmp(path, "/sys", 4) == 0) {
+		up_users();
+		ret = do_sys_readlink(path, buf, size);
+		down_users();
+		return ret;
+	}
+
 	return -EINVAL;
 }
 
@@ -864,7 +918,7 @@ int lxcfs_mkdir(const char *path, mode_t mode)
 	return -EPERM;
 }
 
-#ifdef HAVE_FUSE3
+#if HAVE_FUSE3
 int lxcfs_chown(const char *path, uid_t uid, gid_t gid, struct fuse_file_info *fi)
 #else
 int lxcfs_chown(const char *path, uid_t uid, gid_t gid)
@@ -893,7 +947,7 @@ int lxcfs_chown(const char *path, uid_t uid, gid_t gid)
  * really make sense for cgroups.  So just return 0 always but do
  * nothing.
  */
-#ifdef HAVE_FUSE3
+#if HAVE_FUSE3
 int lxcfs_truncate(const char *path, off_t newsize, struct fuse_file_info *fi)
 #else
 int lxcfs_truncate(const char *path, off_t newsize)
@@ -902,6 +956,9 @@ int lxcfs_truncate(const char *path, off_t newsize)
 	if (strncmp(path, "/cgroup", 7) == 0)
 		return 0;
 
+	if (strncmp(path, "/sys", 4) == 0)
+		return 0;
+
 	return -EPERM;
 }
 
@@ -919,7 +976,7 @@ int lxcfs_rmdir(const char *path)
 	return -EPERM;
 }
 
-#ifdef HAVE_FUSE3
+#if HAVE_FUSE3
 int lxcfs_chmod(const char *path, mode_t mode, struct fuse_file_info *fi)
 #else
 int lxcfs_chmod(const char *path, mode_t mode)
@@ -943,6 +1000,24 @@ int lxcfs_chmod(const char *path, mode_t mode)
 	return -ENOENT;
 }
 
+#if HAVE_FUSE3
+static void *lxcfs_init(struct fuse_conn_info *conn, struct fuse_config *cfg)
+#else
+static void *lxcfs_init(struct fuse_conn_info *conn)
+#endif
+{
+	char *error;
+	void *(*__lxcfs_fuse_init)(struct fuse_conn_info * conn, void * cfg);
+
+	dlerror();
+	__lxcfs_fuse_init = (void *(*)(struct fuse_conn_info * conn, void * cfg))dlsym(dlopen_handle, "lxcfs_fuse_init");
+	error = dlerror();
+	if (error)
+		return log_error(NULL, "%s - Failed to find lxcfs_fuse_init()", error);
+
+	return __lxcfs_fuse_init(conn, NULL);
+}
+
 const struct fuse_operations lxcfs_ops = {
 	.access		= lxcfs_access,
 	.chmod		= lxcfs_chmod,
@@ -950,6 +1025,7 @@ const struct fuse_operations lxcfs_ops = {
 	.flush		= lxcfs_flush,
 	.fsync		= lxcfs_fsync,
 	.getattr	= lxcfs_getattr,
+	.init		= lxcfs_init,
 	.mkdir		= lxcfs_mkdir,
 	.open		= lxcfs_open,
 	.opendir	= lxcfs_opendir,
@@ -960,35 +1036,34 @@ const struct fuse_operations lxcfs_ops = {
 	.rmdir		= lxcfs_rmdir,
 	.truncate	= lxcfs_truncate,
 	.write		= lxcfs_write,
+	.readlink	= lxcfs_readlink,
 
 	.create		= NULL,
 	.destroy	= NULL,
-#ifndef HAVE_FUSE3
+#if !HAVE_FUSE3
 	.fgetattr	= NULL,
 #endif
 	.fsyncdir	= NULL,
-#ifndef HAVE_FUSE3
+#if !HAVE_FUSE3
 	.ftruncate	= NULL,
 	.getdir		= NULL,
 #endif
 	.getxattr	= NULL,
-	.init		= NULL,
 	.link		= NULL,
 	.listxattr	= NULL,
 	.mknod		= NULL,
-	.readlink	= NULL,
 	.rename		= NULL,
 	.removexattr	= NULL,
 	.setxattr	= NULL,
 	.statfs		= NULL,
 	.symlink	= NULL,
 	.unlink		= NULL,
-#ifndef HAVE_FUSE3
+#if !HAVE_FUSE3
 	.utime		= NULL,
 #endif
 };
 
-static void usage()
+static void usage(void)
 {
 	lxcfs_info("Usage: lxcfs <directory>\n");
 	lxcfs_info("lxcfs is a FUSE-based proc, sys and cgroup virtualizing filesystem\n");
@@ -1007,61 +1082,6 @@ static void usage()
 	exit(EXIT_FAILURE);
 }
 
-static inline bool is_help(char *w)
-{
-	return strcmp(w, "-h") == 0 ||
-	       strcmp(w, "--help") == 0 ||
-	       strcmp(w, "-help") == 0 ||
-	       strcmp(w, "help") == 0;
-}
-
-static inline bool is_version(char *w)
-{
-	return strcmp(w, "-v") == 0 ||
-	       strcmp(w, "--version") == 0 ||
-	       strcmp(w, "-version") == 0 ||
-	       strcmp(w, "version") == 0;
-}
-
-static bool swallow_arg(int *argcp, char *argv[], char *which)
-{
-	for (int i = 1; argv[i]; i++) {
-		if (strcmp(argv[i], which) != 0)
-			continue;
-
-		for (; argv[i]; i++)
-			argv[i] = argv[i + 1];
-
-		(*argcp)--;
-		return true;
-	}
-
-	return false;
-}
-
-static bool swallow_option(int *argcp, char *argv[], char *opt, char **v)
-{
-	for (int i = 1; argv[i]; i++) {
-		if (!argv[i + 1])
-			continue;
-
-		if (strcmp(argv[i], opt) != 0)
-			continue;
-
-		do {
-			*v = strdup(argv[i + 1]);
-		} while (!*v);
-
-		for (; argv[i + 1]; i++)
-			argv[i] = argv[i + 2];
-
-		(*argcp) -= 2;
-		return true;
-	}
-
-	return false;
-}
-
 static int set_pidfile(char *pidfile)
 {
 	__do_close int fd = -EBADF;
@@ -1088,7 +1108,7 @@ static int set_pidfile(char *pidfile)
 		return log_error(-1, "Error truncating PID file '%s': %m", pidfile);
 
 	ret = snprintf(buf, sizeof(buf), "%ld\n", (long)getpid());
-	if (ret < 0 || ret >= sizeof(buf))
+	if (ret < 0 || (size_t)ret >= sizeof(buf))
 		return log_error(-1, "Failed to convert pid to string %m");
 
 	if (write(fd, buf, ret) != ret)
@@ -1097,14 +1117,63 @@ static int set_pidfile(char *pidfile)
 	return move_fd(fd);
 }
 
+static const struct option long_options[] = {
+	{"debug",		no_argument,		0,	'd'	},
+	{"disable-swap",	no_argument,		0,	'u'	},
+	{"enable-loadavg",	no_argument,		0,	'l'	},
+	{"foreground",		no_argument,		0,	'f'	},
+	{"help",		no_argument,		0,	'h'	},
+	{"version",		no_argument,		0,	'v'	},
+
+	{"enable-cfs",		no_argument,		0,	  0	},
+	{"enable-pidfd",	no_argument,		0,	  0	},
+
+	{"pidfile",		required_argument,	0,	'p'	},
+	{								},
+};
+
+static int append_comma_separate(char **s, const char *append)
+{
+	int ret;
+	char *news;
+	size_t append_len, len;
+
+	if (!append)
+		return 0;
+
+	append_len = strlen(append);
+	if (!append_len)
+		return 0;
+
+	if (*s) {
+		len = strlen(*s);
+		news = realloc(*s, len + append_len + 2);
+	} else {
+		len = 0;
+		news = realloc(NULL, append_len + 1);
+	}
+	if (!news)
+		return -ENOMEM;
+
+	if (*s)
+		ret = snprintf(news + len, append_len + 2, ",%s", append);
+	else
+		ret = snprintf(news, append_len + 1, "%s", append);
+	if (ret < 0)
+		return -EIO;
+
+	*s = news;
+	return 0;
+}
+
 int main(int argc, char *argv[])
 {
 	int pidfile_fd = -EBADF;
 	int ret = EXIT_FAILURE;
-	char *pidfile = NULL, *saveptr = NULL, *token = NULL, *v = NULL;
+	char *pidfile = NULL, *token = NULL;
 	char pidfile_buf[STRLITERALLEN(RUNTIME_PATH) + STRLITERALLEN("/lxcfs.pid") + 1] = {};
 	bool debug = false, foreground = false;
-#ifndef HAVE_FUSE3
+#if !HAVE_FUSE3
 	bool nonempty = false;
 #endif
 	bool load_use = false;
@@ -1112,8 +1181,12 @@ int main(int argc, char *argv[])
 	 * what we pass to fuse_main is:
 	 * argv[0] -s [-f|-d] -o allow_other,directio argv[1] NULL
 	 */
-	int nargs = 5, cnt = 0;
-	char *newargv[6];
+	int fuse_argc = 0;
+	int c, idx, new_argc;
+	char *fuse_argv[7];
+	const char *fuse_opts = NULL;
+	char *new_fuse_opts = NULL;
+	char *const *new_argv;
 	struct lxcfs_opts *opts;
 
 	opts = malloc(sizeof(struct lxcfs_opts));
@@ -1121,90 +1194,144 @@ int main(int argc, char *argv[])
 		lxcfs_error("Error allocating memory for options");
 		goto out;
 	}
+
 	opts->swap_off = false;
 	opts->use_pidfd = false;
 	opts->use_cfs = false;
+	opts->version = 1;
 
-	/* accomodate older init scripts */
-	swallow_arg(&argc, argv, "-s");
+	while ((c = getopt_long(argc, argv, "dulfhvso:p:", long_options, &idx)) != -1) {
+		switch (c) {
+		case 0:
+			if (strcmp(long_options[idx].name, "enable-pidfd") == 0)
+				opts->use_pidfd = true;
+			else if (strcmp(long_options[idx].name, "enable-cfs") == 0)
+				opts->use_cfs = true;
+			else
+				usage();
+			break;
+		case 'd':
+			debug = true;
+			break;
+		case 'f':
+			foreground = true;
+			break;
+		case 'l':
+			load_use = true;
+			break;
+		case 'o':
+			if (fuse_opts) {
+				lxcfs_error("Specifying -o multiple times is unsupported");
+				usage();
+			}
 
-	/* -f / --foreground */
-	foreground = swallow_arg(&argc, argv, "-f");
-	if (swallow_arg(&argc, argv, "--foreground"))
-		foreground = true;
-
-	/* -d / --debug */
-	debug = swallow_arg(&argc, argv, "-d");
-	if (swallow_arg(&argc, argv, "--debug"))
-		debug = true;
+			fuse_opts = optarg;
+			break;
+		case 'p':
+			pidfile = optarg;
+			break;
+		case 's':
+			/* legacy argument: ignore */
+			break;
+		case 'u':
+			opts->swap_off = true;
+			break;
+		case 'v':
+			lxcfs_info("%s", STRINGIFY(PROJECT_VERSION));
+			exit(EXIT_SUCCESS);
+		default:
+			usage();
+		}
+	}
 
 	if (foreground && debug)
 		log_exit("Both --debug and --forgreound specified");
 
-	/* -l / --enable-loadavg */
-	load_use = swallow_arg(&argc, argv, "-l");
-	if (swallow_arg(&argc, argv, "--enable-loadavg"))
-		load_use = true;
+	new_argv = &argv[optind];
+	new_argc = argc - optind;
 
-	/* -u / --disable-swap */
-	opts->swap_off = swallow_arg(&argc, argv, "-u");
-	if (swallow_arg(&argc, argv, "--disable-swap"))
-		opts->swap_off = true;
-
-	/* --enable-pidfd */
-	opts->use_pidfd = swallow_arg(&argc, argv, "--enable-pidfd");
-
-	/* --enable-cfs */
-	if (swallow_arg(&argc, argv, "--enable-cfs"))
-		opts->use_cfs = true;
-
-	if (swallow_option(&argc, argv, "-o", &v)) {
-		/* Parse multiple values */
-		for (; (token = strtok_r(v, ",", &saveptr)); v = NULL) {
-			if (strcmp(token, "allow_other") == 0) {
-				/* Noop. this is the default. Always enabled. */
-			} else if (strcmp(token, "nonempty") == 0) {
-#ifdef HAVE_FUSE3
-				/* FUSE3: Noop. this is the default. */
-#else
-				nonempty = true;
-#endif
-			} else {
-				lxcfs_error("Warning: unexpected fuse option %s", v);
-				free(v);
-				exit(EXIT_FAILURE);
-			}
-		}
-		free(v);
-		v = NULL;
-	}
-
-	/* -p / --pidfile */
-	if (swallow_option(&argc, argv, "-p", &v))
-		pidfile = v;
-	if (!pidfile && swallow_option(&argc, argv, "--pidfile", &v))
-		pidfile = v;
-
-	if (argc == 2  && is_version(argv[1])) {
-		lxcfs_info("%s", VERSION);
-		exit(EXIT_SUCCESS);
-	}
-
-	if (argc != 2 || is_help(argv[1]))
+	/* Older LXCFS versions printed help when used without any argument. */
+	if (new_argc == 0)
 		usage();
 
-	do_reload();
-	if (install_signal_handler(SIGUSR1, sigusr1_reload)) {
-		lxcfs_error("%s - Failed to install SIGUSR1 signal handler", strerror(errno));
+	if (new_argc != 1) {
+		lxcfs_error("Missing mountpoint");
 		goto out;
 	}
 
-	newargv[cnt++] = argv[0];
+	fuse_argv[fuse_argc++] = argv[0];
 	if (debug)
-		newargv[cnt++] = "-d";
+		fuse_argv[fuse_argc++] = "-d";
 	else
-		newargv[cnt++] = "-f";
-	newargv[cnt++] = "-o";
+		fuse_argv[fuse_argc++] = "-f";
+	fuse_argv[fuse_argc++] = "-o";
+
+	/* Parse additional fuse options. */
+	if (fuse_opts) {
+		char *dup;
+
+		dup = strdup(fuse_opts);
+		if (!dup) {
+			lxcfs_error("Failed to copy fuse options");
+			goto out;
+		}
+
+		lxc_iterate_parts(token, dup, ",") {
+			/* default */
+			if (strcmp(token, "allow_other") == 0)
+				continue;
+
+			/* default for LXCFS */
+			if (strcmp(token, "direct_io") == 0)
+				continue;
+
+			/* default for LXCFS */
+			if (strncmp(token, "entry_timeout", STRLITERALLEN("entry_timeout")) == 0)
+				continue;
+
+			/* default for LXCFS */
+			if (strncmp(token, "attr_timeout", STRLITERALLEN("entry_timeout")) == 0)
+				continue;
+
+			/* default for LXCFS */
+			if (strncmp(token, "allow_other", STRLITERALLEN("allow_other")) == 0)
+				continue;
+
+			/* default with fuse3 */
+			if (strcmp(token, "nonempty") == 0) {
+				#if !HAVE_FUSE3
+				nonempty = true;
+				#endif
+				continue;
+			}
+
+			if (append_comma_separate(&new_fuse_opts, token)) {
+				lxcfs_error("Failed to copy fuse argument \"%s\"", token);
+				free(dup);
+				goto out;
+			}
+		}
+		free(dup);
+	}
+
+	if (append_comma_separate(&new_fuse_opts, "allow_other,entry_timeout=0.5,attr_timeout=0.5")) {
+		lxcfs_error("Failed to copy fuse argument \"allow_other,entry_timeout=0.5,attr_timeout=0.5\"");
+		goto out;
+	}
+
+#if !HAVE_FUSE3
+	if (nonempty) {
+		if (append_comma_separate(&new_fuse_opts, "nonempty")) {
+			lxcfs_error("Failed to copy fuse argument \"nonempty\"");
+			goto out;
+		}
+	}
+
+	if (append_comma_separate(&new_fuse_opts, "direct_io")) {
+		lxcfs_error("Failed to copy fuse argument \"nonempty\"");
+		goto out;
+	}
+#endif
 
 	/*
 	 * We can't use default_permissions since we still support systems that
@@ -1215,16 +1342,16 @@ int main(int argc, char *argv[])
 	 * shouldn't guarantee that we don't need more complicated access
 	 * helpers for proc and sys virtualization in the future.
 	 */
-#ifdef HAVE_FUSE3
-	newargv[cnt++] = "allow_other,entry_timeout=0.5,attr_timeout=0.5";
-#else
-	if (nonempty)
-		newargv[cnt++] = "allow_other,direct_io,entry_timeout=0.5,attr_timeout=0.5,nonempty";
-	else
-		newargv[cnt++] = "allow_other,direct_io,entry_timeout=0.5,attr_timeout=0.5";
-#endif
-	newargv[cnt++] = argv[1];
-	newargv[cnt++] = NULL;
+
+	fuse_argv[fuse_argc++] = new_fuse_opts;
+	fuse_argv[fuse_argc++] = new_argv[0];
+	fuse_argv[fuse_argc] = NULL;
+
+	do_reload();
+	if (install_signal_handler(SIGUSR1, sigusr1_reload)) {
+		lxcfs_error("%s - Failed to install SIGUSR1 signal handler", strerror(errno));
+		goto out;
+	}
 
 	if (!pidfile) {
 		snprintf(pidfile_buf, sizeof(pidfile_buf), "%s/lxcfs.pid", RUNTIME_PATH);
@@ -1238,7 +1365,7 @@ int main(int argc, char *argv[])
 	if (load_use && start_loadavg() != 0)
 		goto out;
 
-	if (!fuse_main(nargs, newargv, &lxcfs_ops, opts))
+	if (!fuse_main(fuse_argc, fuse_argv, &lxcfs_ops, opts))
 		ret = EXIT_SUCCESS;
 
 	if (load_use)
@@ -1249,6 +1376,7 @@ out:
 		dlclose(dlopen_handle);
 	if (pidfile)
 		unlink(pidfile);
+	free(new_fuse_opts);
 	free(opts);
 	close_prot_errno_disarm(pidfile_fd);
 	exit(ret);
