@@ -3,24 +3,16 @@
 #ifndef __LXCFS_PROC_FUSE_H
 #define __LXCFS_PROC_FUSE_H
 
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE
-#endif
+#include "config.h"
 
-#ifndef FUSE_USE_VERSION
-#define FUSE_USE_VERSION 26
-#endif
+#include "lxcfs_fuse.h"
 
-#define _FILE_OFFSET_BITS 64
-
-#include <fuse.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
 
-#include "config.h"
 #include "macro.h"
 
 __visible extern int proc_getattr(const char *path, struct stat *sb);
